#include "local.h"
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include "son.h"

#ifdef ALLOC_H
#include <alloc.h>
#endif

#ifdef MALLOC_H
#include <malloc.h>
#endif

#ifdef NO_MALLOC_H
/* 
** malloc.h is not defined on the Apollo machines 
*/
	extern char *malloc(unsigned);
	extern int free(char *);
	extern char *remalloc(char *,unsigned);
	extern char *calloc(unsigned,unsigned);
	extern int cfree(char *);
	extern char *memalign(unsigned,unsigned);
	extern char *valloc(unsigned);

#endif /* NO_MALLOC_H */


#define NEW_CLEAR 0		
#define NEW_TRACE 0

/******************************************************************
NEW_DBG 

	Determines if memory allocation is good.
	Make sure that the next location after the allocated area is 
	not change.  Keeps trace of:
		1. size of allocated memory
		2. a uniq number which is incremented every time an allocation
		   is done, including new, rnew and cnew.  May be used to 
		   trace when an allocation is done.  To do so break in
		   "memory_nu" when alloc_nu == the allocation number.
		3. File and line where allocation was done.
		4. Address of memory being allocated.
		5. Total amount of memory being allocated.

NEW_TRACE

	Prints out all allocations.
	where n = the uniq. allocation number.

NEW_CLEAR

	Make sure that uninitialize memory is not used.  

	new():  For a new() set allocation memory allocated to FILL.
	rnew(): Create new memory using a new().  Set memory to FILL, then copy 
		over the previous memory.  Next set original memory to FILL,
		then free the allocated memory.
	free(): Set all memory to FILL then free it.

	SET IN sonlib.h
NOTE:
	1. Production code should have all debugging turned off.
	2. To use NEW_CLEAR or NEW_TRACE must have NEW_DBG turn on.

Debug routines:
	To print out all allocations call print_new(-1)
	To print out a specific allocation call print_new(n)

******************************************************************/

#if NEW_CLEAR | NEW_TRACE
#if ! NEW_DBG

	ERROR: To use NEW_CLEAR or NEW_TRACE, must have NEW_DBG also set

#endif
#endif

#if NEW_DBG

#undef new
#undef rnew
#undef cnew
#undef sFree
#undef stop_if_nomem

#endif

#ifdef DOS
#define ADDR	long int
#else
#define ADDR	char *
#endif

static int total = 0;
static int stop_on_nomem = 1;
static int make_core = 0;

#if NEW_DBG

#define END_CHAR	'X'
#define FILL		'Y'

char *bsearch();
void qsort();
static int alloc_nu = 1;
static char **ptr_tbl=NULL;
static int *ptr_size=NULL;
static int *ptr_nu=NULL;
static char **ptr_file=NULL;
static int *ptr_linenu=NULL;
static int nel=0;
static char *filename = "unknown";
static int line_number = 0;

#if NEW_TRACE
static int trace = NEW_TRACE;
#endif

static int cmp_ptr(
#if NeedFunctionPrototypes
	char **x,
	char **y
#endif
);

static void ptr_tbl_add(
#if NeedFunctionPrototypes
	char *x,
	int size
#endif
);

static void
ptr_sort(
#if NeedFunctionPrototypes
	void
#endif
);

static void
print_new(
#if NeedFunctionPrototypes
	int x
#endif
);

static
void memory_nu(
#if NeedFunctionPrototypes
	int alloc_nu
#endif
);
#endif	/* End: NEW_DBG */


char *new(size) unsigned size;
/*************************************************************/
/* Allocate memory with malloc, if error exit program.       */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11mar88 */ 
{ char *A;

#if NEW_DBG
	++size;
#endif

  if ( (A=malloc(size)) == NULL ) 
  { if ( stop_on_nomem ) 
    { fprintf(stderr,"Out of memory (malloc).\n");
      if ( make_core ) {
	core_program();
      }
      exit_program(1);
    } 
    else { 
#if NEW_TRACE
      if (trace>0) printf("new: %s:%d size=%d, OUT OF MEMORY\n",
			filename,line_number,size);
#endif
      return (NULL);
    }
  }


#if NEW_DBG

  ptr_check();

  A[--size] = END_CHAR;
  total += size;

#if NEW_TRACE
	if (trace>0) printf("new: %s:%d size=%d, addr=%lx\n",
			filename,line_number,size,A);
#endif

#if NEW_CLEAR
	memset(A,FILL,size);
#endif

  ptr_tbl_add(A,size);
#endif

  return(A);
} /*new*/

char *rnew(ptr,size) char *ptr;  unsigned size;
/*************************************************************/
/* Allocate memory with malloc, if error exit program.       */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11mar88 */ 
{ char *A;

#if NEW_DBG
  int i;
  char *tbl;
  char **loc;
  int next;
  int num;

   ++size;
  ptr_check();
#endif

#if NEW_CLEAR
  A=malloc(size);
#else
  A=(char *)realloc(ptr,size);
#endif

  if ( A == NULL )
  { if ( stop_on_nomem ) 
    { fprintf(stderr,"Out of memory (realloc).\n");
      if ( make_core ) {
	core_program();
      }
      exit_program(1);
    } 
    else { 
#if NEW_TRACE
      if (trace>0) printf("rnew: %s:%d size=%d, OUT OF MEMORY\n",
			filename,line_number,size);
#endif
      return (NULL);
    }
  }

#if NEW_DBG



#if NEW_TRACE
	if (trace>0) printf("rnew: %s:%d size=%d oldaddr=%lx addr=%lx\n",
			filename,line_number,size,ptr,A);
#endif
  
  loc = (char **)bsearch((char *)(&ptr),(char *)ptr_tbl,
				      nel,sizeof(ptr),cmp_ptr);
  next = (int)(loc - ptr_tbl);
  total += (size-ptr_size[next]-1);

  if (loc==NULL) {
    printf("rnew: size=%d oldaddr=%lx addr=%lx\n",size,ptr,A);
    print_new(-1);
    announce_bug("rnew");
  }
  *loc = A;

#if NEW_CLEAR
  memset(A,FILL,size);
  if ( size < ptr_size[next] ) {
	num = size;
  } else {
	num = ptr_size[next];
  }
  memcpy(A,ptr,num);
  memset(ptr,FILL,ptr_size[next]);
  free(ptr);
#endif
  A[--size] = END_CHAR;

  tbl = ptr_tbl[next];

  free(ptr_file[next]);
  for ( i = next; i < nel-1;  ++i ) {
    ptr_tbl[i] = ptr_tbl[i+1];
    ptr_size[i] = ptr_size[i+1];
    ptr_nu[i] = ptr_nu[i+1];
    ptr_file[i] = ptr_file[i+1];
    ptr_linenu[i] = ptr_linenu[i+1];
  }
  memory_nu(alloc_nu);
  ptr_tbl[nel-1] = tbl;
  ptr_size[nel-1] = size;
  ptr_file[nel-1] = malloc((strlen(filename)+1)*sizeof(char));
  strcpy( ptr_file[nel-1],filename);
  ptr_linenu[nel-1] = line_number;
  ptr_nu[nel-1] = alloc_nu++;

  ptr_sort();

  ptr_check();

#endif

  return(A);
} /*rnew*/

char *cnew(n,size) unsigned n,size;
/*************************************************************/
/* Allocate memory with malloc, if error exit program.       */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr11mar88 */ 
{ char *A;
#if NEW_DBG
	++size;
#endif
  if ( (A=(char *)calloc(n,size)) == NULL ) 
  { if ( stop_on_nomem ) 
    { fprintf(stderr,"Out of memory (calloc).\n");
      if ( make_core ) {
	core_program();
      }
      exit_program(1);
    } 
    else { 
#if NEW_TRACE
      if (trace>0) printf("cnew: %s:%d size=%d, OUT OF MEMORY\n",
			filename,line_number,size*n);
#endif
      return (NULL);
    }
  }

#if NEW_DBG

  --size;
  ptr_check();

#if NEW_TRACE
	if (trace>0) printf("cnew: %s:%d size=%d  addr=%lx\n",
			filename,line_number,size*n,A);
#endif

  A[n*size] = END_CHAR;
  total += (n*size);

  /* Bug checking tbl */
  ptr_tbl_add(A,n*size);

#endif

  return(A);
} /*cnew*/

void sFree(ptr) char *ptr;
/*************************************************************/
/* Release the memory pointed to by ptr.                     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr16may89 */ 
{
#if NEW_DBG
  char **loc,**loc_stop;
  int *iloc,*iloc_stop;
  char **cloc,**cloc_stop;
  int next;
  
#if NEW_TRACE
	if (trace>0) printf("free: %s:%d addr=%lx\n",filename,line_number,ptr);
#endif

    ptr_check();

    loc = (char **)bsearch((char *)(&ptr),(char *)ptr_tbl,
				       nel,sizeof(ptr),cmp_ptr);
    if (loc==NULL) {
	print_new(-1);
	printf("free: %s:%d addr=%lx\n",filename,line_number,ptr);
	loc = (char **)bsearch((char *)(&ptr),(char *)ptr_tbl,
			       nel,sizeof(ptr),cmp_ptr);
	announce_bug("sFree");
    }
    next = (int)(loc - ptr_tbl);
    total -= ptr_size[next];

#if NEW_CLEAR
    /* Clear out the memory */
    memset(ptr,FILL,ptr_size[next]);
#endif

    for (loc_stop=ptr_tbl+nel-1; loc<loc_stop; loc++) {
	*loc = *(loc+1);
    }

    iloc = ptr_size + next;
    for (iloc_stop=ptr_size+nel-1; iloc<iloc_stop; iloc++) {
	*iloc = *(iloc+1);
    }

    cloc = ptr_file + next;
    for (cloc_stop=ptr_file+nel-1; cloc<cloc_stop; cloc++) {
	*cloc = *(cloc+1);
    }

    iloc = ptr_linenu + next;
    for (iloc_stop=ptr_linenu+nel-1; iloc<iloc_stop; iloc++) {
	*iloc = *(iloc+1);
    }
    iloc = ptr_nu + next;
    for (iloc_stop=ptr_nu+nel-1; iloc<iloc_stop; iloc++) {
	*iloc = *(iloc+1);
    }
    nel--;


#if NEW_TRACE
	if (trace>1) print_new(-1);
#endif

#endif
  

  free(ptr);
  return;
} /*sFree*/

void
stop_if_nomem(value)
int value;
/* krs25may91 */ 
/*************************************************************/
/* Determines if new(), rnew() and cnew() should terminate   */
/* program if no memory is available.  If value is TRUE then */
/* terminate program, else return NULL			     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
{
	stop_on_nomem = value;

#if NEW_TRACE
	if (trace>0) printf("stop_if_nomem: %s:%d %s\n",
			filename,line_number, (value)? "On" : "Off" );
#endif
}

void
new_core(value) 
int value;
/* krs25may91 */ 
/*************************************************************/
/* If value is TRUE then produce a core file when out of     */
/* memory occurs.					     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
{
	make_core = value;
}

void
core_program()
/* krs25may91 */ 
/*************************************************************/
/* Produce a core file					     */
/* memory occurs.					     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
{
	int a,b;
	a=1;

	/* Make sure we are not stopped by the optimizer compiler */
	for (b = -1; b< 1; ++b ) {
		if (b > 0 ) printf("%d\n",a/b);
	}
}

#if NEW_DBG
static int cmp_ptr(x,y)  char **x,**y;
/*************************************************************/
/* Compare two character pointers.                           */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr16may89 */ 
{
  if (*x == *y) return(0);
  if ( (ADDR)*x > (ADDR)*y ) {
	return 1;
  } else {
  	return -1;
  }
} /*cmp_ptr*/

static void ptr_tbl_add(x,size)  char *x; int size;
/*************************************************************/
/* Add a character pointer to the table.                     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr16may89 */ 
{ static int max=0;

  if (nel>=max)
  { max += 256;
    ptr_tbl = (ptr_tbl==NULL) ? (char **)malloc(max*sizeof(*ptr_tbl))
			      : (char **)realloc(ptr_tbl,max*sizeof(*ptr_tbl));
    ptr_size = (ptr_size==NULL) ? (int *)malloc(max*sizeof(int))
			      : (int *)realloc(ptr_size,max*sizeof(int));
    ptr_file = (ptr_file==NULL) ? (char **)malloc(max*sizeof(char *))
			      : (char **)realloc(ptr_file,max*sizeof(char *));
    ptr_linenu = (ptr_linenu==NULL) ? (int *)malloc(max*sizeof(int))
			      : (int *)realloc(ptr_linenu,max*sizeof(int));
    ptr_nu = (ptr_nu==NULL) ? (int *)malloc(max*sizeof(int))
			      : (int *)realloc(ptr_nu,max*sizeof(int));
  }

  memory_nu(alloc_nu);
  *(ptr_tbl + nel) = x;
  *(ptr_size + nel) = size;
  *(ptr_file + nel) = malloc((strlen(filename)+1)*sizeof(char));
  strcpy( *(ptr_file+nel),filename);
  *(ptr_linenu + nel) = line_number;
  *(ptr_nu + nel) = alloc_nu++;
  nel++;
  ptr_sort();

} /*ptr_tbl_add*/

static void
ptr_sort()
{
	int i,next;
	char *tbl;
	int size;
	int nu;
	int linenu;
	char *file;

	next = nel - 1;
	tbl = ptr_tbl[next];
	size = ptr_size[next];
	file = ptr_file[next];
	linenu = ptr_linenu[next];
	nu = ptr_nu[next];

	for ( i=0; i < nel-1; ++i ) {
		if ( (ADDR)ptr_tbl[i] >  (ADDR)tbl ) {
			next =i;
			break;
		}
	}
	for ( i = nel-1; i > next; --i ) {
		ptr_tbl[i] = ptr_tbl[i-1];
		ptr_size[i] = ptr_size[i-1];
		ptr_file[i] = ptr_file[i-1];
		ptr_linenu[i] = ptr_linenu[i-1];
		ptr_nu[i] = ptr_nu[i-1];
	}
	ptr_tbl[next] = tbl;
	ptr_size[next] = size;
	ptr_file[next] = file;
	ptr_linenu[next] = linenu;
	ptr_nu[next] = nu;

#if NEW_TRACE
	if (trace>1) print_new(-1);
#endif
}

static void
print_new(x)
int x;
{
	int found;
	int i;

	printf("Index  File          line  alloc_nu size   address\n");
	found = 0;
	for (i=0; i < nel; ++i) {
		if ( x == -1 || ptr_nu[i] == x ) {
			printf("%4d %+15s:%6d %5d  %5d   %lx\n",
				i,ptr_file[i],ptr_linenu[i],
				ptr_nu[i],ptr_size[i],ptr_tbl[i] );
		found = 1;
		}
	}
	if ( x == -1 ) {
		printf("\n");
		printf("Total: %d\n",total);
	} else {
		if ( ! found ) {
			printf("Not found\n");
		}
	}
}

void
ptr_check()
{
	int i;
	int error;
	error = 0;
	for (i=0; i < nel; ++i) {
		if ( *(ptr_tbl[i]+ptr_size[i]) != END_CHAR ) {
			fprintf(stderr,"CORRUPT MEMORY: i=%5d  %s:%d nu=%5d size=%5d   addr=%lx - %lx\n",
				i,ptr_file[i],ptr_linenu[i],
				ptr_nu[i],ptr_size[i],ptr_tbl[i],
				ptr_tbl[i] +ptr_size[i] );
			fprintf(stderr,"CORRUPT MEMORY: last added = %d\n",alloc_nu-1);
			error = -1;
		}
	}
	if ( error ) {
		if ( make_core ) {
			core_program();
		}
		exit_program (1);
	}
}


#endif

/*
#ifdef DOS
char *
bsearch(key, base, nelem, width, fcmp)
void *key;
void *base;
size_t nelem;
size_t width;
int (*fcmp)();
{
	size_t low, high, mid;
	int rc;

	low = 0;
	high = nelem;
	while ( low <= high ) {
		mid = (low+high)/2;
		rc = fcmp( (char *)key, (char *)base+mid*width);
		if ( rc == 0 ) return( (char *)base+mid*width );
		if ( rc < 0 ) high = mid - 1;
		else          low  = mid + 1;
	}
	return (NULL);
}
#endif
*/

#if NEW_DBG


char *
new_tr( file, linenu, size)
char *file;
int linenu;
unsigned size;
{
	char *A;
	filename = file;
	line_number = linenu;
	A = new(size);
	filename = "unknown";
	line_number = 0;

	return (A);
}

char *
rnew_tr( file, linenu, ptr, size)
char *file;
int linenu;
char *ptr;
unsigned size;
{
	char *A;

	filename = file;
	line_number = linenu;
	A = rnew(ptr,size);
	filename = "unknown";
	line_number = 0;
	return (A);
}

char *cnew_tr(file, linenu, n,size) 
char *file;
int linenu;
unsigned n,size;
{
	char *A;

	filename = file;
	line_number = linenu;
	A = cnew(n,size);
	filename = "unknown";
	line_number = 0;
	return (A);
}

void
sFree_tr(file, linenu, ptr) 
char *file;
int linenu;
char *ptr;
{
	filename = file;
	line_number = linenu;
	sFree(ptr);
	filename = "unknown";
	line_number = 0;
}

void
stop_if_nomem_tr(file, linenu, value)
char *file;
int linenu;
int value;
{
	filename = file;
	line_number = linenu;
	stop_if_nomem(value);
	filename = "unknown";
	line_number = 0;
}

static
void memory_nu(alloc_nu)
int alloc_nu;
{
	if ( alloc_nu < -1 ) {
		fprintf(stderr,"Internal Error - alloc_nu = 5d\n",alloc_nu);
	}
}

#endif

void
print_memory_alloc()
{
	printf("Memory Allocated: %d\n",total);
}

void
print_all_new()
{
#if NEW_DBG
	print_new(-1);
#endif
}
